#!/usr/bin/perl -w
#
# Take a GPX tracklog and give some stats about speed, time, distance.
#
# Copyright (c) R. Odom <russ@gloomytrousers.co.uk> 2008
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Instructions:
#	Takes 1 arg: the GPX file to analyse
#
# Latest version available from
# http://www.gloomytrousers.co.uk/open_source/gps_tracklog_stats.shtml
# Comments, suggestions for improvements, etc. welcome!
#
# Version history:
#	v0.1.0	01 Mar 2008
#		Initial version. Some useful stats but quick 'n' dirty in places.
#

# Config values:
my $distunit = 'mile';	# Any unit supported by Geo::Distance
my $timeunit = 'hour';	# The time unit to display
my $timeunitmult = 3600;	# How many seconds in $timeunit
my $movethreshold = 0.2;	# Min speed at which we are considered to be moving in $distunit per $timeunit (should be big enough to allow for drift due to GPS inaccuracy when stopped)
my $sataccurate = 7;	# Number of satelites needed to consider a point accurate when calculating max speed	# TODO: look at fix accuracy too, if present in GPX
my $maxspeedminduration = 30;	# Max speed is calculated over a collection of points spanning at least this many seconds
###TODO: my $maxspeedmintravel = 0.2;	# Max speed is calculated over a collection of points covering at least this distance

use Geo::Gpx;	# Tracklog parsing/iteration.
use Geo::Distance;	# Distance calculation in all sorts of wierd and wonderful units
use Geo::Functions qw(dms_deg);	# Some handy conversions
use POSIX qw(strftime);

# Program version info
$VER = '0.1.0';
$LONGVER = '$Date: 2008-03-01 20:16:24 +0000 (Sat, 01 Mar 2008) $, $Revision: 162 $';	# SVN keywords
$LONGVER =~ s/\s*\$//g;
$LONGVER =~ s/\s*\(.+\)\s*,/, /g;

print "gpxstats ver $VER ($LONGVER)\n\n";

# Load the tracklog
open (GPX, $ARGV[0]) or die("Can't open GPX file $ARGV[0]\n");
### TODO: load XML direct from filehandle instead: my $gpx = Geo::Gpx->new( input => GPX );
### I'm doing summat wrong - getting: Can't call method "get_token" on an undefined value at /usr/lib/perl5/site_perl/5.8.8/XML/Descent.pm line 54
$xml = '';
while (<GPX>) { $xml .= $_; }
my $gpx = Geo::Gpx->new( xml => $xml );
close GPX;

# Distance calculator object
# TODO: Factor usage of $distcalc->distance() into a sub that takes 2 points as args
my $distcalc = new Geo::Distance();

my $totaldist = 0, $totaltime = 0, $pointcount = 0, $satcount = 0, $satmin = 0, $satmax = 0;
my $maxspeed = 0, @speedcheckpoints = ();
my ($firstpt, $prevpt);

# Process the points
### TODO: ideally should iterate over the track segments separately
my $iter = $gpx->iterate_trackpoints();
while (my $pt = $iter->()) {

	# TODO: some way of determining points that are of dubious accuracy and skipping them
	# TODO: account for the fact that most stats are optional in the GPX spec

	# Some running stats
	$pointcount++;
	$satcount += $pt->{sat};
	$satmax = $pt->{sat} if ($pt->{sat} > $satmax);
	###print "Point: $pt->{lat}, $pt->{lon}, $pt->{time}, $pt->{ele}\n";

	push @speedcheckpoints, $pt;	# Add this point to the set we check for max speed

	if (defined($firstpt)) {	# No calculation can be done if we're on first point

		# Calculate stats for this point
		# TODO: there is no vertical distance component in the calculation here. Remember that's always in metres.
		my $thistime = ($pt->{time} - $prevpt->{time})/$timeunitmult;
		next if (! $thistime);
		my $thisdist = $distcalc->distance($distunit, $prevpt->{lon},$prevpt->{lat} => $pt->{lon},$pt->{lat});
		my $thisspeed = $thisdist/$thistime;	# TODO: use extensions->rmc:speed if present

		# Update running total
		$totaldist += $thisdist;
		$totaltime += $thistime;

		# Have we moved?
		if ($thisspeed > $movethreshold) {
			$movingdist += $thisdist;
			$movingtime += $thistime;
		}

		# Check for new max speed...
		# See if we can throw away the earliest point because the remainder span enough time
		shift @speedcheckpoints while ($#speedcheckpoints >= 1 && $speedcheckpoints[$#speedcheckpoints]->{time} - $speedcheckpoints[1]->{time} >= $maxspeedminduration);

		# See if our points have enough satellites to consider 'em accurate
		my $ok = 1;
		foreach my $sppoint (@speedcheckpoints) {
			if ($sppoint->{sat} < $sataccurate) {
				$ok = 0;
				last;
			}
		}
		if ($ok) {	# Enough accuracy
			my $spdist = 0, $sptime = 0, $spspeed = 0;
			for ($i = 1; $i <= $#speedcheckpoints; $i++) {
				$sptime += $speedcheckpoints[$i]->{time} - $speedcheckpoints[$i-1]->{time};
				$spdist += $distcalc->distance($distunit, $speedcheckpoints[$i-1]->{lon},$speedcheckpoints[$i-1]->{lat} => $speedcheckpoints[$i]->{lon},$speedcheckpoints[$i]->{lat});
			}
			$sptime /= $timeunitmult;
			$spspeed = $spdist/$sptime;
			if ($spdist/$sptime > $maxspeed && $spdist/$sptime >= $movethreshold) {
				# New max speed found
				# TODO: 'my' these:
				$maxspeed = $spspeed;
				$maxspeedduration = $sptime;
				$maxspeeddist = $totaldist;
				$maxspeedpt1 = $speedcheckpoints[0];
				$maxspeedpt2 = $speedcheckpoints[$#speedcheckpoints];
				$maxspeednumpoints = $#speedcheckpoints + 1;
			}
		}
	} else {
		# First point
		$firstpt = $pt;
		$satmin = $pt->{sat};
		###print "Stats available: " . (join ', ', (sort keys %{$pt})) . "\n\n";
		# TODO: can we get any useful info from the extensions? Seems to be just the values as a string, no XML tags so pretty useless
	}

	$prevpt = $pt;
}

# Output statistics
my $crowfliesdist = $distcalc->distance($distunit, $firstpt->{lon},$firstpt->{lat} => $prevpt->{lon},$prevpt->{lat});
printf("Trip start:	%s at %s\n", fmt_coords_dms($firstpt->{lon}, $firstpt->{lat}), fmt_secs_datetime($firstpt->{time}));
printf("Trip end:	%s at %s\n", fmt_coords_dms($prevpt->{lon}, $prevpt->{lat}), fmt_secs_datetime($prevpt->{time}));
printf("As crow flies:	%0.2f %ss from start to end\n\n", $crowfliesdist, $distunit);

printf("Trip overall:	%0.2f %ss in %s\n", $totaldist, $distunit, fmt_secs_timespan($totaltime*$timeunitmult));
printf("Average speed:	%0.2f %ss/%s\n", $totaldist/$totaltime, $distunit, $timeunit);
if ($maxspeed) {
	printf("Maximum speed:	%0.2f %ss/%s over %0.2f %ss in %s (%d track points)\n", $maxspeed, $distunit, $timeunit, $maxspeed*$maxspeedduration, $distunit, fmt_secs_timespan($maxspeedduration*$timeunitmult), $maxspeednumpoints);
	printf("		at %0.2f %ss, around %s from start\n", $maxspeeddist, $distunit, fmt_secs_timespan($maxspeedpt1->{time} - $firstpt->{time}));
	printf("		between %s, %s with %d satellites\n", fmt_coords_dms($maxspeedpt1->{lon}, $maxspeedpt1->{lat}), fmt_secs_time($maxspeedpt1->{time}), $maxspeedpt1->{sat});
	printf("		and     %s, %s with %d satellites\n\n", fmt_coords_dms($maxspeedpt2->{lon}, $maxspeedpt2->{lat}), fmt_secs_time($maxspeedpt2->{time}), $maxspeedpt2->{sat});
} else {
	print "Maximum speed:	None recorded\n\n";
}

printf("Whilst moving:	%0.2f %ss in %s (moving faster than %s %ss/%s)\n", $movingdist, $distunit, fmt_secs_timespan($movingtime*$timeunitmult), $movethreshold, $distunit, $timeunit);
printf("Average speed:	%0.2f %ss/%s\n", $movingdist/$movingtime, $distunit, $timeunit);
printf("Time moving:	%s (%0.1f%%)\n", fmt_secs_timespan($movingtime*$timeunitmult), 100*$movingtime/$totaltime);
printf("Time stopped:	%s (%0.1f%%)\n\n", fmt_secs_timespan(($totaltime - $movingtime)*$timeunitmult), 100*($totaltime - $movingtime)/$totaltime);

printf("Route eff'cy:	%0.2f%% (from start to end in a straight line is 100%% efficient)\n\n", 100*$crowfliesdist/$totaldist);

printf("Track points:	%d total, an of average %0.1f per %s or 1 every %s\n", $pointcount, $pointcount/$totaltime, $timeunit, fmt_secs_timespan($timeunitmult*$totaltime/$pointcount));
printf("Sats used:	%d min, %d max, %0.1f average\n\n", $satmin, $satmax, $satcount/$pointcount);

# Takes 2 numbers: longitude and latitude
# Returns nice human-readable co-ord string
sub fmt_coords_dms {
	my ($lond, $lonm, $lons, $lonns) = dms_deg(pop);
	my ($latd, $latm, $lats, $latew) = dms_deg(pop);
	return sprintf("%02d\xc2\xB0%02d'%02.2f\"%s, %02d\xc2\xB0%02d'%02.2f\"%s", $lond, $lonm, $lons, ($lonns>0?'N':'S'), $latd, $latm, $lats, ($latew>0?'E':'W'));
}

# Takes a Unix time (in seconds since epoch)
# Returns nice human-readable date/time string
sub fmt_secs_datetime {
	my $time = shift;
	return sprintf("%s %s", fmt_secs_date($time), fmt_secs_time($time))
}

# Takes a Unix time (in seconds since epoch)
# Returns nice human-readable date string
sub fmt_secs_date {
	return strftime("%d %b %Y", localtime(shift));
}

# Takes a Unix time (in seconds since epoch)
# Returns nice human-readable time string (no date part)
sub fmt_secs_time {
	return strftime("%T", localtime(shift));
}

# Takes a timespan in secs
# Returns nice human-readable time in hh:mm:ss format
sub fmt_secs_timespan {
	#TODO: trips longer than a day
	return strftime("%T", gmtime(shift));
}
